import abc
import collections
import datetime
import lxml.etree
import mimetypes
import posixpath
import uuid
import zipfile

from .utils import E
from .utils import getxmlattr
from .utils import NS


VERSIONS = {}

# this isn't set up correctly in the alpine image ; maybe it should be a package?
mimetypes.add_type('application/pls+xml', '.pls')
mimetypes.add_type('application/smil+xml', '.smil')
mimetypes.add_type('application/x-dtbncx+xml', '.ncx')
mimetypes.add_type('application/xhtml+xml', '.htm')
mimetypes.add_type('application/xhtml+xml', '.html')
mimetypes.add_type('application/xhtml+xml', '.xhtml')
mimetypes.add_type('application/xhtml+xml', '.xml')
mimetypes.add_type('font/otf', '.otf')
mimetypes.add_type('font/ttf', '.ttf')
mimetypes.add_type('font/woff', '.woff')

class Epub(abc.ABC):
	version = None

	_default_toc_path = None

	def __init_subclass__(cls, *args, **kwargs):
		super().__init_subclass__(*args, **kwargs)
		VERSIONS[cls.version] = cls

	def __init__(self, zf, opfpath):
		self._opfpath = opfpath
		self._zf = zf

		self.cover = None
		self.landmarks = Landmarks()
		self.layout = Layout()
		self.manifest = Manifest()
		self.spine = Spine()
		self.toc = Toc(None, None)
		self.meta = {
			'contributors': [], # list of AS with role and file-as
			'creators': [], # list of AS with role and file-as
			'dates': {'creation': None, 'publication': None, 'modification': None},
			'description': None, # AS with lang
			'identifiers': [], # list of AS with id and scheme
			'publisher': None, # AS with lang
			'languages': [],
			'source': None,
			'subjects': [],
			'titles': [], # list of AS with lang
		}

	def _init_read(self, opftree):
		self._read_manifest(opftree)
		self._read_meta(opftree)
		self._read_spine(opftree)
		self._read_toc(opftree)

		uid_id = opftree.get('unique-identifier')
		self.uid = next(filter(lambda i: i.get('id') == uid_id, self.meta['identifiers']), None)

	def _init_write(self):
		self.uid = AttributedString(str(uuid.uuid4()))
		self.uid['id'] = 'uid_id'
		self.uid['scheme'] = 'uuid'
		self.meta['identifiers'] = [self.uid]
		self.meta['languages'] = [AttributedString('und')]

		self.toc.item = self.manifest.add(self._default_toc_path)

		self._writestr('mimetype', b'application/epub+zip', compress_type=zipfile.ZIP_STORED)

		container = E['container'].container(
			{'version': '1.0'},
			E['container'].rootfiles(
				E['container'].rootfile({
					'full-path': self._opfpath,
					'media-type': 'application/oebps-package+xml',
				}),
			),
		)
		self._writestr(
			'META-INF/container.xml',
			lxml.etree.tostring(container, pretty_print=True),
			compress_type=zipfile.ZIP_STORED,
		)

	def _write_opf(self):
		self.meta['dates']['modification'] = datetime.datetime.now(datetime.timezone.utc)

		if self.toc.item is not None:
			self._write_toc()

		pkg = self._xml_opf()
		self._writestr(self._opfpath, lxml.etree.tostring(pkg, pretty_print=True))

	@abc.abstractmethod
	def _read_toc(self, opftree): # pragma: no cover
		...

	def _append_to_toc(self, toc, href, title, children):
		# href is relative to self.toc.item
		href = self._from_toc_href(href)
		toc_item = toc.append(href, title)
		for href, title, c in children:
			self._append_to_toc(toc_item.children, href, title, c)

	@abc.abstractmethod
	def _read_meta(self, opftree): # pragma: no cover
		...

	def _read_manifest(self, opftree):
		for item in opftree.findall('./opf:manifest/opf:item', NS):
			self.manifest[getxmlattr(item, 'id')] = getxmlattr(item, 'href')

	def _read_spine(self, opftree):
		for item in opftree.findall('./opf:spine/opf:itemref', NS):
			item = self.manifest[getxmlattr(item, 'idref')]
			self.spine.append(item)

	@abc.abstractmethod
	def _write_toc(self): # pragma: no cover
		...

	def _from_toc_href(self, href):
		root = posixpath.dirname(self.toc.item.href)
		opfdir = posixpath.dirname(self._opfpath)

		href = posixpath.join(opfdir, root, href)
		href = posixpath.relpath(href, opfdir)
		href = posixpath.normpath(href)

		return href

	def _to_toc_href(self, href):
		# href is relative to the opfdir
		# return an href suitable for use in the toc

		opfdir = posixpath.dirname(self._opfpath)
		tocdir = posixpath.dirname(posixpath.join(opfdir, self.toc.item.href))

		href = posixpath.join(opfdir, href)
		href = posixpath.relpath(href, tocdir)
		href = posixpath.normpath(href)

		return href

	def _xml_meta(self):
		return E['opf'].metadata(E['dc'].format('application/epub+zip'))

	def _xml_manifest_item(self, item):
		return E['opf'].item({
			'id': item.iid,
			'href': item.href,
			'media-type': item.mimetype or 'application/octet-stream',
		})

	def _xml_manifest(self):
		return E['opf'].manifest(*(
			self._xml_manifest_item(item)
			for item in sorted(self.manifest.values(), key=lambda it: it.iid)
		))

	def _xml_spine_item(self, item):
		return E['opf'].itemref({'idref': item.iid})

	def _xml_spine(self):
		return E['opf'].spine(*(self._xml_spine_item(item) for item in self.spine))

	def _xml_opf(self):
		return E['opf'].package(
			{
				'version': self.version,
				'unique-identifier': self.uid['id'],
				'prefix': 'rendition: http://www.idpf.org/vocab/rendition/#',
			},
			self._xml_meta(),
			self._xml_manifest(),
			self._xml_spine(),
		)

	def write(self, *args, **kwargs):
		raise NotImplementedError('Use writestr')

	def _writestr(self, *args, **kwargs):
		self._zf.writestr(*args, **kwargs)

	def writestr(self, item, data, **kwargs):
		if isinstance(item, zipfile.ZipInfo):
			raise NotImplementedError('item should be a path relative to the opfdir or an Item')
		if not isinstance(item, self.manifest.Item) or item.iid not in self.manifest:
			item = self.manifest.add(item)

		self._writestr(self.__opfpath(item.href), data, **kwargs)
		return item

	def open(self, item, *args, **kwargs):
		if isinstance(item, self.manifest.Item):
			item = item.href
		return self._zf.open(self.__opfpath(item), *args, **kwargs)

	@property
	def isbn(self):
		for i in self.meta['identifiers']:
			scheme = i.get('scheme', '').lower()
			val = str(i)
			if scheme == 'isbn':
				return i
			if scheme == '' and len(val) == 13 and val[:3] in ('978', '979'):
				return i
		raise KeyError('Unknown ISBN')

	def __opfpath(self, path):
		path = posixpath.join(posixpath.dirname(self._opfpath), path)
		path = posixpath.normpath(path)
		return path

	def __repr__(self):
		return '<Epub {} (len(manifest): {}, len(spine): {})>'.format(self.version, len(self.manifest), len(self.spine))


class Manifest(dict):
	class Item:
		def __init__(self, iid, href):
			self.iid = iid
			self.href = href

		@property
		def mimetype(self):
			return mimetypes.guess_type(self.href)[0]

		def __repr__(self):
			return '<Manifest.Item {}>'.format(self.__dict__)

	def add(self, item):
		if not isinstance(item, self.Item):
			item = self.Item('item-{}'.format(len(self)), item)
		self[item.iid] = item
		return item

	def __setitem__(self, k, v):
		if isinstance(v, str):
			v = self.Item(k, v)
		if not isinstance(v, self.Item):
			raise TypeError('The manifest needs to be a dict of Manifest.Item')
		super().__setitem__(k, v)

	def byhref(self, href):
		href = href.split('#', 1)[0]
		it = filter(lambda item: item.href == href, self.values())
		try: return next(it)
		except StopIteration: raise KeyError(href)


class Spine(list):
	def append(self, item):
		if not isinstance(item, Manifest.Item):
			raise TypeError('The spine needs to be a list of Manifest.Item')
		return super().append(item)


class TocItems(list):
	class Item:
		def __init__(self, href, title):
			self.href = href # href is the canonical href, relative to the OPF dir
			self.title = title
			self.children = TocItems()

		def __repr__(self):
			return '<Toc.Item {}>'.format(self.__dict__)

	def append(self, item, title=None, children=None):
		if isinstance(item, str):
			if title is None:
				raise TypeError('Need a title to add an href to the TOC')
			item = self.Item(item, title)
			for a in children or []:
				item.children.append(*a)
		if not isinstance(item, self.Item):
			raise TypeError('The TOC needs to be a list of Toc.Item')
		super().append(item)
		return item


class Toc(TocItems):
	def __init__(self, item, title):
		self.item = item
		self.title = title
		super().__init__()


class AttributedString(collections.UserDict):
	def __init__(self, value, **kwargs):
		self.value = value
		self.data = kwargs

	def __bool__(self):
		return bool(self.value)

	def __str__(self):
		return self.value

	def __repr__(self):
		return '<AttributedString {!r} {}>'.format(self.value, self.data)


class Landmarks(list):
	class Item:
		def __init__(self, *, kind, href, title):
			self.kind = kind
			self.href = href
			self.title = title

		def __repr__(self):
			return '<Landmarks.Item {}>'.format(self.__dict__)

	def append(self, item=None, **kwargs):
		if item is None:
			item = self.Item(**kwargs)
		if not isinstance(item, self.Item):
			raise TypeError('The landmarks needs to be a list of Landmarks.Item')
		super().append(item)
		return item


class Layout(dict):
	class Item:
		pass

	class Fixed(Item):
		def __init__(self, *, orientation: str = None, spread: str = None, page_spread: str = None):
			self.orientation = orientation
			self.spread = spread
			self.page_spread = page_spread

		def _copy(self):
			return type(self)(orientation=self.orientation, spread=self.spread, page_spread=self.page_spread)

		def __repr__(self):
			return '<Layout.Fixed {}>'.format(self.__dict__)

	class Reflowable(Item):
		def _copy(self):
			return type(self)()

		def __repr__(self):
			return '<Layout.Reflowable>'

	def __init__(self):
		self.default = self.Reflowable()
		super().__init__()

	def __getitem__(self, key):
		if key not in self:
			self[key] = self.default._copy()
		return super().__getitem__(key)

	def __setitem__(self, key, val):
		if not isinstance(val, self.Item):
			raise TypeError('The layout items needs to be Layout.Item elements')
		super().__setitem__(key, val)



import lxml.etree
import posixpath
import zipfile

from .epub import VERSIONS
from .utils import NS


def _opf_from_metainf(zf):
	try:
		zi = zf.getinfo('META-INF/container.xml')
	except KeyError:
		return

	with zf.open(zi) as f:
		tree = lxml.etree.parse(f)
	rootfile = tree.find('./container:rootfiles/container:rootfile', NS)
	if rootfile is None:
		return

	path = rootfile.get('full-path')
	try:
		return zf.getinfo(path)
	except KeyError:
		return

def _find_opf_zi(zf):
	# adapted from https://github.com/kovidgoyal/calibre/blob/68febe94ca2baf2a0979668b7b61a1e3b0432f0f/src/calibre/ebooks/conversion/plugins/epub_input.py#L233
	# the opf path should be specified in the META-INF/container.xml file, but we try to handle gracefully broken epubs
	zi = _opf_from_metainf(zf)
	if zi is not None:
		return zi

	for zi in zf.infolist():
		if all((
			zi.filename.lower().endswith('.opf'),
			'__MACOSX' not in zi.filename,
			not posixpath.basename(zi.filename).startswith('.'),
		)):
			return zi

	raise ValueError('Not a valid ePub file: could not find the OPF file')

class open:
	def __init__(self, infile, mode='r', version=None, opfpath=None):
		if mode not in ('r', 'w'):
			raise TypeError('Supported modes are r and w')
		if mode == 'r' and opfpath is not None:
			raise TypeError('opfpath should only be used in w mode')
		if mode == 'w' and version is None:
			raise TypeError('version is required in w mode')

		self._mode = mode
		self._opfpath = opfpath
		self._version = version
		self._zf = zipfile.ZipFile(infile, mode=mode)


	def __enter__(self):
		self._zf.__enter__()

		if self._mode == 'r':
			opfzi = _find_opf_zi(self._zf)
			with self._zf.open(opfzi) as f:
				opftree = lxml.etree.parse(f).getroot()

			version = self._version or opftree.get('version')

			self._epub = VERSIONS[version](self._zf, opfzi.filename)
			self._epub._init_read(opftree)

		else:
			assert self._mode == 'w'
			opfpath = self._opfpath or 'content.opf'
			self._epub = VERSIONS[self._version](self._zf, opfpath)
			self._epub._init_write()

		return self._epub

	def __exit__(self, *args):
		if self._mode == 'w':
			self._epub._write_opf()
		self._zf.__exit__(*args)
		del self._epub
		del self._zf



import itertools
import logging
import lxml.etree

from .epub import AttributedString
from .epub import Epub
from .utils import E
from .utils import format_date
from .utils import getxmlattr
from .utils import ns
from .utils import NS
from .utils import parse_date


logger = logging.getLogger(__name__)

class Epub20(Epub):
	version = '2.0'
	_default_toc_path = 'toc.ncx'

	def _read_toc(self, opftree):
		toc_id = getxmlattr(opftree.find('./opf:spine', NS), 'toc')
		if toc_id is None:
			return

		def parse(tag):
			for np in tag.findall('./ncx:navPoint', NS):
				content = np.find('./ncx:content', NS)
				if content is None:
					continue
				text = np.find('./ncx:navLabel/ncx:text', NS)
				yield (
					getxmlattr(content, 'src'),
					(text.text if text is not None else None) or '',
					parse(np),
				)

		self.toc.item = self.manifest[toc_id]
		with self.open(self.toc.item) as f:
			ncx = lxml.etree.parse(f).getroot()

		for a in parse(ncx.find('./ncx:navMap', NS)):
			self._append_to_toc(self.toc, *a)

		title_tag = ncx.find('./ncx:docTitle/ncx:text', NS)
		if title_tag is not None:
			self.toc.title = title_tag.text or ''

	__meta = [
		# tag, attributes, multiple
		('title', ('xml:lang',), True),
		('creator', ('opf:role', 'opf:file-as'), True),
		('subject', (), True),
		('description', (), False),
		('publisher', (), False),
		('contributor', ('opf:role', 'opf:file-as'), True),
		# Drop type
		# Drop format
		# date handled manually
		('identifier', ('id', 'opf:scheme'), True),
		('source', (), False),
		('language', (), True),
		# Drop relation
		# Drop coverage
		# Drop rights
	]
	def _read_meta(self, opftree):
		metadata = opftree.find('./opf:metadata', NS)
		def extract(tag, attrs, ns='dc'):
			for t in metadata.findall('{}:{}'.format(ns, tag), NS):
				yield AttributedString(t.text or '', **{
					k.split(':', 1)[-1]: getxmlattr(t, k)
					for k in attrs
					if getxmlattr(t, k) is not None
				})

		for tag, attrs, multi in self.__meta:
			f = list if multi else lambda d: next(d, None)
			self.meta[tag + ('s' if multi else '')] = f(extract(tag, attrs))

		for astr in extract('date', ('opf:event',)):
			if 'event' not in astr:
				continue
			if astr['event'] in self.meta['dates']:
				self.meta['dates'][astr['event']] = parse_date(astr)

		for astr in extract('meta', ('name', 'content'), ns='opf'):
			if astr.get('name') == 'cover':
				try:
					cover = self.manifest[astr['content']]
				except KeyError:
					logger.debug('Cover not found: %s', astr['content'])
				else:
					self.cover = cover

		guide = opftree.find('./opf:guide', NS)
		if guide is not None:
			for t in guide.findall('opf:reference', NS):
				self.landmarks.append(
					kind=getxmlattr(t, 'type'),
					title=getxmlattr(t, 'title'),
					href=getxmlattr(t, 'href'),
				)

	def _xml_meta(self):
		for k, li in (('Default', self.layout.default), *self.layout.items()):
			if not isinstance(li, self.layout.Reflowable):
				raise ValueError('ePub 2.0 only support reflowable layouts. Invalid layout {}: {!r}'.format(k, li))

		meta = super()._xml_meta()

		for k, v in self.meta['dates'].items():
			if v is not None:
				meta.append(E['dc'].date(format_date(v), {ns('opf:event'): k}))

		for tag, attrs, multi in self.__meta:
			todo = self.meta.get(tag + ('s' if multi else ''))
			if not todo:
				continue
			if not multi:
				todo = [todo]
			for astr in todo:
				tag = getattr(E['dc'], tag)(str(astr))
				for k in attrs:
					val = astr.get(k.split(':', 1)[-1])
					if val:
						tag.attrib[ns(k)] = val
				meta.append(tag)

		if self.cover is not None:
			meta.append(E['opf'].meta({'name': 'cover', 'content': self.cover.iid}))

		return meta

	def _xml_spine(self):
		spine = super()._xml_spine()
		if self.toc.item is not None:
			spine.attrib['toc'] = self.toc.item.iid
		return spine

	def _xml_opf(self):
		res = super()._xml_opf()
		if self.landmarks:
			res.append(E['opf'].guide(*(
				E['opf'].reference({'type': it.kind, 'title': it.title, 'href': it.href})
				for it in self.landmarks
			)))
		return res

	def _write_toc(self):
		ids = itertools.count()

		def depth(toc):
			if not toc:
				return 0
			return max(depth(item.children) for item in toc) + 1

		def navpoints(toc):
			for item in toc:
				np = E['ncx'].navPoint(
					{'id': 'np-{}'.format(next(ids))},
					E['ncx'].navLabel(E['ncx'].text(item.title)),
					E['ncx'].content({'src': self._to_toc_href(item.href)}),
				)
				if item.children:
					for c in navpoints(item.children):
						np.append(c)
				yield np

		toc = E['ncx'].ncx(
			{'version': '2005-1'},
			E['ncx'].head(
				E['ncx'].meta({'name': 'dtb:uid', 'content': str(self.uid)}),
				E['ncx'].meta({'name': 'dtb:depth', 'content': str(depth(self.toc))}),
			),
			E['ncx'].docTitle(E['ncx'].text(self.toc.title or '')),
			E['ncx'].navMap(*navpoints(self.toc)),
		)

		data = lxml.etree.tostring(toc, pretty_print=True)
		self.writestr(self.toc.item, data)



#!/usr/bin/env python

import re
import setuptools


requirements = [
	'lxml >= 4.3.4, < 5',
]

setuptools.setup(
	name='dawn',
	version='0.9.0',
	description='Open and modify ePubs',
	author='Arthur Darcet',
	author_email='hello+dawn@glose.com',
	url='http://git.glose.com/opensource/dawn',
	license='MIT',
	packages=['dawn'],
	install_requires=requirements,
	setup_requires=[
		'pytest-runner >= 5.1, < 6',
	],
	tests_require=requirements + [
		'pytest',
		'pytest-cov',
	],
	test_suite='test',
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules',
	],
)
